class StageParams:
    PARAM_MD5 = "md5"
    PARAM_CMD = "cmd"
    PARAM_WDIR = "wdir"
    PARAM_DEPS = "deps"
    PARAM_OUTS = "outs"
    PARAM_LOCKED = "locked"  # backward compatibility
    PARAM_FROZEN = "frozen"
    PARAM_META = "meta"
    PARAM_ALWAYS_CHANGED = "always_changed"
    PARAM_PARAMS = "params"
    PARAM_METRICS = "metrics"
    PARAM_PLOTS = "plots"
    PARAM_DESC = "desc"
